# scanf for Ruby, v.1.1.0
#
# August 28, 2002
#
# A product of the Austin Ruby Codefest (Austin, Texas, August 2002)
#
# Copyright (c) 2002, David Alan Black
# Released under the same licensing terms as Ruby itself.
#
# See README for information on usage and details of license.
#
#

module Scanf

  class FormatSpecifier < String

    attr_reader :re_string, :string

    private

    def skip;  /^\s*%\*/.match(self); end
    def to_i(s); s.to_i if s &&! skip; end
    def to_f(s); s.to_f if s &&! skip; end
    def to_h(s); s.hex if s &&! skip; end
    def to_o(s); s.oct if s &&! skip; end
    def to_n(s); Integer(s) if s &&! skip; end
    def keep(s); s unless skip; end
    def nil_proc(s); nil; end

    public

    def count_space?
      /(?:\A|\S)%\*?\d*c|\[/.match(self)
    end

    def initialize(*args)
      super(*args)

      h = '[A-Fa-f0-9]'

      @re_string, @handler = 
	case self

	  # %[[:...:]]
	when /%\*?(\[\[:[a-z]+:\]\])/
	  [ "(#{$1}+)", :keep ]

	  # %5[[:...:]]
	when /%\*?(\d+)(\[\[:[a-z]+:\]\])/
	  [ "(#{$2}{1,#{$1}})", :keep ]

	  # %[...]
	when /%\*?\[([^\]]*)\]/
	  yes = $1
	  if /^\^/.match(yes) then no = yes[1..-1] else no = '^' + yes end
	  [ "([#{yes}]+)(?=[#{no}]|\\z)", :keep ]

	  # %5[...]
	when /%\*?(\d+)\[([^\]]*)\]/
	  yes = $2
	  w = $1
	  [ "([#{yes}]{1,#{w}})", :keep ]

	  # %i
	when /%\*?i/
	  [ "([-+]?(?:(?:0[0-7]+)|(?:0[Xx]#{h}+)|(?:[1-9]\\d+)))", :to_n ]

	  # %5i
	when /%\*?(\d+)i/
	  n = $1.to_i
	  s = "("
	  if n > 1 then s += "[1-9]\\d{1,#{n-1}}|" end
	  if n > 1 then s += "0[0-7]{1,#{n-1}}|" end
	  if n > 2 then s += "[-+]0[0-7]{1,#{n-2}}|" end
	  if n > 2 then s += "[-+][1-9]\\d{1,#{n-2}}|" end
	  if n > 2 then s += "0[Xx]#{h}{1,#{n-2}}|" end
	  if n > 3 then s += "[-+]0[Xx]#{h}{1,#{n-3}}|" end
	  s += "\\d"
	  s += ")"
	  [ s, :to_n ]

	  # %d, %u
	when /%\*?[du]/
	  [ '([-+]?\d+)', :to_i ]

	  # %5d, %5u
	when /%\*?(\d+)[du]/
	  n = $1.to_i
	  s = "("
	  if n > 1 then s += "[-+]\\d{1,#{n-1}}|" end
	  s += "\\d{1,#{$1}})"
	  [ s,:to_i ]

	  # %x
	when /%\*?[Xx]/
	  [ "([-+]?(?:0[Xx])?#{h}+)", :to_h ]

	  # %5x
	when /%\*?(\d+)[Xx]/
	  n = $1.to_i
	  s = "("
	  if n > 3 then s += "[-+]0[Xx]#{h}{1,#{n-3}}|" end
	  if n > 2 then s += "0[Xx]#{h}{1,#{n-2}}|" end
	  if n > 1 then s += "[-+]#{h}{1,#{n-1}}|" end
	  s += "#{h}{1,#{n}}"
	  s += ")"
	  [ s, :to_h ]

	  # %o
	when /%\*?o/
	  [ '([-+]?[0-7]+)', :to_o ]

	  # %5o
	when /%\*?(\d+)o/
	  [ "([-+][0-7]{1,#{$1.to_i-1}}|[0-7]{1,#{$1}})", :to_o ]

	  # %f
	when /%\*?f/
	  [ '([-+]?((\d+(?>(?=[^\d.]|$)))|(\d*(\.(\d*([eE][-+]?\d+)?)))))', :to_f ]

	  # %5f
	when /%\*?(\d+)f/
	  [ "(\\S{1,#{$1}})", :to_f ]

	  # %5s
	when /%\*?(\d+)s/
	  [ "(\\S{1,#{$1}})", :keep ]

	  # %s
	when /%\*?s/
	  [ '(\S+)', :keep ]

	  # %c
	when /\s%\*?c/
	  [ "\\s*(.)", :keep ]

	  # %c
	when /%\*?c/
	  [ "(.)", :keep ]

	  # %5c (whitespace issues are handled by the count_*_space? methods)
	when /%\*?(\d+)c/
	  [ "(.{1,#{$1}})", :keep ]

	  # %%
	when /%%/
	  [ '(\s*%)', :nil_proc ]

	  # literal characters
	else
	  [ "(#{Regexp.escape(self)})", :nil_proc ]
	end

      @re_string = '\A' + @re_string
    end

    def cook(str)
      send(@handler, str)
    end

    def to_re
      Regexp.new(@re_string,Regexp::MULTILINE)
    end

    def match(str)
      s = str.dup
      s.sub!(/\A\s+/,'') unless count_space?
      m = to_re.match(s)
      @string = m.to_a[1]
      m
    end

    def letter
      /%\*?\d*([a-z\[])/.match(self).to_a[1]
    end

    def width
      w = /%\*?(\d+)/.match(self).to_a[1]
      w && w.to_i || nil
    end

  end

  class ScanfResult < Array
    attr_accessor :format_string, :chars_matched, :chars_left, :string_left
  end

  class FormatString < Array

    attr_reader :eos, :eofmt, :latest, :index, :failed, :mcount, :pos, :space

    SPECIFIERS = 'diuXxofeEgsc'
    REGEX = /
        # possible space, followed by...
          (?:\s*
          # percent sign, followed by...
	    %
	    # another percent sign, or...
	      (?:%|
		 # optional assignment suppression flag
		 \*?
		 # optional maximum field width
		 \d*
		   # named character class, ...
		   (?:\[\[:\w+:\]\]|
		   # traditional character class, or...
		      \[[^\]]*\]|
		   # specifier letter.
		      [#{SPECIFIERS}])))|
	    # or miscellaneous characters
	      [^%\s]+/ix

    def initialize(str)
      s = str.to_s
      return unless /\S/.match(s)
      @space = true if /\s\z/.match(s)
      replace s.scan(REGEX).map {|spec| FormatSpecifier.new(spec) }
    end

    def prune(n=mcount)
      n.times { shift }
    end

    def all_but_one?
      mcount == size - 1
    end

    def match(str)
      accum = []
      @pos = 0
      @eos = str.dup
      @mcount = 0

      each_with_index do |spec,i|

	# Save some state:

	# Whatever happens, for now this spec is the latest:
	@latest = spec

	# Is this spec the last one in the format string?
	@eofmt = (i == size - 1)
	
	m = spec.match(str)
	@failed = !m
	break if @failed

	# Update state:

	# What's left of the string:
	@eos = m.post_match

	# Number of specs that have successfully matched:
	@mcount += 1
	
	accum << spec.cook(m[1])
	ss = str.size
	str = @eos
	@pos += ss - str.size

	# Gobble up trailing white space if appropriate
	if @space && @eofmt
	  str = str.sub(/(\s*)\z/,'')
	  @pos += $1.size if $1
	end
	break if str.empty?
      end
      accum.compact
    end
  end
end

class IO

  def scanf(fstr,&b)
    return block_scanf(fstr,&b) if b
    skan = []
    return skan unless fstr.size > 0

    matched = 0
    start = pos
    interim = []
    buffer = ""
    fs = Scanf::FormatString.new(fstr)

    loop do
      if eof
	skan.concat(interim)
	matched += buffer.size
	break
      end

      buffer << gets
      m = fs.match(buffer)
      sp = fs.latest

# Match fails
#
# just one spec left?
#   no: fatal failure
#   yes:
#     fs.eos (remaining string) all spaces?
#       yes: does latest spec care about input space?
#         yes: fatal failre
#         no: save interim results and continue

      if fs.failed
	if fs.all_but_one? &&! /\S/.match(fs.eos)
	  if sp.count_space?
	    break
	  else
	    interim.replace(m)
	    next
	  end
	end

# Match succeeds (non-emptily)
#
# is the latest matched spec a '[' or 'c' with a width?
#   yes: is latest.string.size < available width?
#     yes: save interim results
#     no: width is used up, so move on (next)
#   no: is it a '[' with no width?
#   yes: evidently nothing violated it yet, so store
#        interim results and continue (next)

      elsif !m.empty?
	siz = sp.string.to_s.size
	wid = sp.width.to_i
	
	cc_no_width    =   sp.letter == '[' &&! sp.width
	c_or_cc_width  =   sp.letter == 'c' || sp.letter == '[' && sp.width
	c_or_cc_open   =   c_or_cc_width && (siz < wid) &&! eof

	if c_or_cc_open
	  next
	elsif cc_no_width
	  interim.replace(m)
	  next
	end
      end

      skan.concat(m)
      s = buffer.size
      buffer.replace(fs.eos)
      matched += s - buffer.size
      break if fs.eofmt
      fs.prune
    end

    seek(matched + start, IO::SEEK_SET) rescue Errno::ESPIPE

    # mop up remaining space, if we're matching trailing space.
    # (like FormatString#match does for strings)

    if fs.eofmt && fs.space
      c = getc until eof || /\S/.match(c.to_i.chr)
      ungetc(c) if c
    end

    skan
  end

  def block_scanf(fstr)
    fs = Scanf::FormatString.new(fstr)
    all = []
    this = self.scanf(fs)
    begin
      all.push(yield this) unless this.empty?
      this = self.scanf(fs)
    end until this.empty? || eof
    all
  end
end

class String

  def scanf(fstr,&b)
    if b
      block_scanf(fstr,&b)
    else
      fs = 
	if fstr.is_a? Scanf::FormatString
	  fstr 
	else 
	  Scanf::FormatString.new(fstr)
	end
      fs.match(self)
    end
  end

  def block_scanf(fstr,&b)
    fs = Scanf::FormatString.new(fstr)
    str = self
    all = []
    this = str.scanf(fs)
    begin
      all.push(yield this) unless this.empty?
      str = fs.eos
      this = str.scanf(fs)
    end until this.empty? || str.empty?
    all
  end
end

module Kernel
  def scanf(fs)
    STDIN.scanf(fs)
  end
end
